// This is OpenHashTable.h


#ifndef OPEN_HASH_TABLE_H
#define OPEN_HASH_TABLE_H


#include "LinkedList.h"


//
// support class for the OpenHashTable class
//

template <class Type>
class OpenHashTableEntry {
	public:
		OpenHashTableEntry(void* key, int keyLength, Type& item);
		~OpenHashTableEntry();

		void* _key;
		int _keyLength;
		Type _item;
};






//
// This is a simple open hash table.  It supports Open Hashing.  If there is a
// collision, multiple entries will be entered into a linked list off the end
// of each entry.
//

template <class Type>
class OpenHashTable {
	public:
		//
		// constructor.  Size should be prime for best performance. The 
		// maxKeyLength should at least be a gues to the maximum key
		// size in bytes.  It is used to tweak the hashing function.
		//
		OpenHashTable(int maxKeyLength, int numItems=5003);
		OpenHashTable(const OpenHashTable<Type>& data);
		~OpenHashTable();


		//
		// insert an item into the hash table.  It will be indexed
		// according to the key that is given.
		//
		void insertItem(void* key, int keyLength, Type& item);


		//
		// see if an item with the given key is in the hash table.
		// if it is in the table, the desired item will be returned.
		// If it is not in the table, NULL will be returned.
		//
		Type getItem(void* key, int keyLength);


		inline int getNumberOfElements()     { return _numItems; }


		//
		// clear the hash table.  This completely removes all entries in
		// the table.
		//
		void resetHashTable();


		OpenHashTable<Type>& operator=(const OpenHashTable<Type>& data);
		//inline Type& operator[](int index);



	protected:
		int _maxKeyLength;
		unsigned int _numItems;

		//
		// we need an array of linked lists...one for each hash table entry
		//
		LinkedList<OpenHashTableEntry<Type>*> *_tableEntries;

		void _reallocateData(int numItems);


		//
		// this is the hashing function.  It will return the position in the
		// hash table where the specified key is (or should be).  The 
		// isPresentInTable variable is also a return value.  It will get
		// set to 1 if the value is already in the table, 0 if it is not
		// presently there.
		//
		int _hashFunction(void* key, int keyLength, int* isPresentInTable);


		//
		// this function will see if the given key is in the linked list
		// that is associated with each field in the hash table.
		//
		int _isInField(int hashTablePos, void* key, int keyLength);


	private:
};







#include <stdlib.h>



//
// The whole impelementation of the open hash table class has to be in
// this header file to satisfy the Microsoft compiler.  Don't you 
// just love Microsoft?
//

#ifndef MIN
#define MIN(a, b) ((a)<(b)?(a):(b))
#endif



template<class Type>
OpenHashTable<Type>::OpenHashTable(int maxKeyLength, int numItems)
	: _maxKeyLength(maxKeyLength),
	  _numItems(0),
	  _tableEntries(NULL) {
	_reallocateData(numItems);
}




template<class Type>
OpenHashTable<Type>::OpenHashTable(const OpenHashTable& data)
	: _maxKeyLength(0),
	  _numItems(0),
	  _tableEntries(NULL) {
	*this = data;
}




template<class Type>
OpenHashTable<Type>::~OpenHashTable() {
	if (_tableEntries) {
		resetHashTable();
		delete [] _tableEntries;
	}
}







//
// insert an item into the hash table.  It will be indexed
// according to the key that is given.
//
template<class Type>
void OpenHashTable<Type>::insertItem(void* key, int keyLength, Type& item) {
	int isPresentInTable;
	int position = _hashFunction(key, keyLength, &isPresentInTable);

	if (!isPresentInTable) {
		OpenHashTableEntry<Type>* entry = new OpenHashTableEntry<Type>(key, keyLength, item);
		_tableEntries[position].appendItem(entry);
	}
}







//
// see if an item with the given key is in the hash table.
// if it is in the table, the desired item will be returned.
// If it is not in the table, NULL will be returned.
//
template<class Type>
Type OpenHashTable<Type>::getItem(void* key, int keyLength) {
	int isPresentInTable;
	int position = _hashFunction(key, keyLength, &isPresentInTable);


	if (isPresentInTable) {
		_tableEntries[position].getHead();
		int entryLen = _tableEntries[position].getLength();


		for (int i=0; i < entryLen; i++) {
			OpenHashTableEntry<Type>* entry = _tableEntries[position].getCurrentItem();

			if (keyLength == entry->_keyLength && 
				memcmp(entry->_key, key, keyLength) == 0) {
				return entry->_item;
			}

			++(_tableEntries[position]);
		}
	}

	return NULL;
}







//
// clear the hash table.  This completely removes all entries in
// the table.
//
template<class Type>
void OpenHashTable<Type>::resetHashTable() {
	if (_tableEntries) {
		for (unsigned int i=0; i < _numItems; i++) {
			int entryLen = _tableEntries[i].getLength();
			for (int j=0; j < entryLen; j++) {
				OpenHashTableEntry<Type>* entry = _tableEntries[i].getTail();
				_tableEntries[i].removeTail();
				delete entry;
			}
		}
	}
}







template<class Type>
OpenHashTable<Type>& 
OpenHashTable<Type>::operator=(const OpenHashTable& data) {
	_reallocateData(data._numItems);

	if (_numItems > 0) {
		for (int i=0; i < _numItems; i++) {
			_tableEntries[i].getHead();
			data._tableEntries[i].getHead();

			for (int j=0; j < data._tableEntries[i].getLength(); j++) {
				_tableEntries[i].setCurrentItem(data._tableEntries[i].getCurrentItem());

				//
				// traverse down the list
				//
				++(_tableEntries[i]);
				++(data._tableEntries[i]);
			}
		}
	}

	_maxKeyLength = data._maxKeyLength;

	return *this;
}




#if 0
template<class Type>
inline Type& OpenHashTable<Type>::operator[](int index) {
	if (index >= _numItems) {
		_reallocateData(MIN(index*2, index+1000));
	}

	if (index >= _maxItemNumberReferenced) {
		_maxItemNumberReferenced = index+1;
	}

	return _data[index];
}
#endif






template<class Type>
void OpenHashTable<Type>::_reallocateData(int numItems) {
	LinkedList<OpenHashTableEntry<Type>*> *oldData = _tableEntries;
	unsigned int oldNumItems = _numItems;

	_numItems = numItems;

	if (numItems > 0) {
		_tableEntries = new LinkedList<OpenHashTableEntry<Type>*>[_numItems];
	}
	else {
		_numItems = 0;
		_tableEntries = NULL;
	}


	unsigned int endOfLoop = MIN(oldNumItems, _numItems);

	for (unsigned int i=0; i < endOfLoop; i++) {
		oldData[i].getHead();

		for (int j=0; j < oldData[i].getLength(); j++) {
			OpenHashTableEntry<Type>* hashEntry = oldData[i].getCurrentItem();
			_tableEntries[i].appendItem(hashEntry);

			//
			// traverse down the list
			//
			++(oldData[i]);
		}
	}


	if (oldData) {
		delete [] oldData;
	}
}








//
// this is the hashing function.  It will return the position in the
// hash table where the specified key is (or should be).  The 
// isPresentInTable variable is also a return value.  It will get
// set to 1 if the value is already in the table, 0 if it is not
// presently there.
//
template<class Type>
int OpenHashTable<Type>::_hashFunction(void* key, int keyLength, 
								       int* isPresentInTable) {
	int position;
	unsigned int hashVal=0;


	//
	// the basic goal of this hashing function is to evenly distribute the
	// items in the hash table in a reproducable manner.
	//

	//
	// for our hashing function, we will shift the bytes over to evenly fill up
	// an unsigned integer.  As we fill, we will xor the bits together to make
	// things even more evenly distributed.
	//
	unsigned char* charKey = (unsigned char*)key;

#if 0
	for (int i=0; i < keyLength; i++) {
		int shift = 8 * (i % 4);
		unsigned int tempKeyVal = (unsigned int)(*charKey) << shift;
		charKey++;
		//
		// not done yet
		//
	}
#endif

	//
	// actually, we will be using this hash fuction.  It comes out of 
	// Data Structures and Algorithm Analysis in C++ by Mark Allen Weiss.
	//
	for (int i=0; i < keyLength; i++) {
		hashVal = (hashVal << 5) + *charKey++;
	}

	position = hashVal % _numItems;

	*isPresentInTable = _isInField(position, key, keyLength);

	return position;
}






//
// this function will see if the given key is in the linked list
// that is associated with each field in the hash table.
//
template<class Type>
int OpenHashTable<Type>::_isInField(int hashTablePos, void* key, int keyLength) {
	int retVal = 0;

	_tableEntries[hashTablePos].getHead();
	int entryLen = _tableEntries[hashTablePos].getLength();


	for (int i=0; i < entryLen; i++) {
		OpenHashTableEntry<Type>* entry = _tableEntries[hashTablePos].getCurrentItem();

		if (keyLength == entry->_keyLength && 
			memcmp(entry->_key, key, keyLength) == 0) {
			retVal = 1;
			break;
		}

		++(_tableEntries[hashTablePos]);
	}

	return retVal;
}











//
// functions for the OpenHashTableEntry support class.
//
template <class Type>
OpenHashTableEntry<Type>::OpenHashTableEntry(void* key, int keyLength, 
											 Type& item)
	: _key(NULL),
	  _keyLength(keyLength),
	  _item(item) {

	if (_keyLength) {
		_key = (void*)(new unsigned char[_keyLength]);
		memcpy(_key, key, _keyLength);
	}
}





template <class Type>
OpenHashTableEntry<Type>::~OpenHashTableEntry() {
	if (_key) {
		delete [] _key;
	}
}




#endif    // OPEN_HASH_TABLE_H


