// This is Image.h
//
// Copyright 1997, 1998     Mark Manca
//


#ifndef __IMAGE
#define __IMAGE

#ifdef WIN32
#include <windows.h>
#endif

#include <GL/gl.h>

#include "Color.h"



class Image {
	public:
		// constructors, Destructors, etc...
		Image();
		~Image();


		GLenum getImageType();
		GLenum getImageDataType();
		int getExtraBytesPerRow();
		void *getImage();

		int getNumColorMapEntries();
		GLenum getColormapDataType();
		void *getColorMap();



		void createNewImage(int width, int height, GLenum type,
		                    GLenum dataType=GL_UNSIGNED_BYTE,
		                    Color backgroundColor=Color(0, 0, 0, 0));

		void convertToRGBA();
		void convertToRGB();


		//
		// these two functions will return a color indexed image (each of the
		// indexes will map to a color table, each of whose entries is a 
		// gray scale value).
		//
		Image* getIntensity();     // get the intensity of the image
		Image* getHue();           // get the hue of the image

		void clear();



		int getDepth();	       // get the depth of the image (in bits per pixel)
		int getNumChannels();


		//
		// set the scale that will be used for drawing the image.  These
		// functions do not effect the image in any other way.
		//
		void setScale(float scale);
		void setXScale(float scale);
		void setYScale(float scale);


		//
		// if you want to set both the width and the height, call
		// set size. it is *much* more efficient because it allocates
		// memory and copies the current image array to the new image array
		// in one pass instead of two separate passes.
		//
		// These methods should only be called after a gl context has
		// been created and made current because they need to figure out
		// how many extra bytes they should allocate.
		//
		void setSize(int width, int height);
		void setWidth(int width) { setSize(width, _height); }
		void setHeight(int height) { setSize(_width, height); }


		float getXScale();
		float getYScale();
		void getSize(int *width, int *height);
		int getWidth();
		int getHeight();
		

		virtual void draw();		// draw the entire image
		virtual void drawRow(int row);	// draw a row of the image
		virtual void draw(int x, int y, int width, int height);

		virtual void drawRed();		// draw the red component of the image
		virtual void drawGreen();	// draw the green component of the image
		virtual void drawBlue();	// draw the blue component
		virtual void drawAlpha();	// draw the alpha component (as a b and w image)

		//
		// reads the pixels from the current glcontext and puts them in the 
		// _imageData array;
		//
		void update();
		void update(int startX, int startY, int width, int height);
		


		int setName(const char *imageName);
		const char *getName();


		//
		// remember that the bottom left pixel is stored in [0].  This function
		// will *only* work if this is an RGBA image.
		//
		Color& operator[](int pixelIndex);

		Image& operator=(const Image& image);




	protected:
		int _width;
		int _height;

		float _xScale;
		float _yScale;

		char *_imageName;

		GLenum _imageType;
		GLenum _imageDataType;

		int _numChannels;
		int _extraRowWidth;

		//
		// the image data is stored the way that OpenGL likes it...
		// ie it stores the bottom left pixel in the very first position
		//
		void *_imageData;

		int  _colormapSize;
		GLenum _colormapDataType;
		void *_colormapData;


		void _allocateImageData();
		void _allocateColormapData();


	private:
};

#endif		// __IMAGE
