// This is ImageFile.cpp
//
// Copyright 1997, 1998     Mark Manca
//

#ifdef WIN32
#include <windows.h>
#endif

#include <GL/gl.h>
#include <string.h>
#include <stdio.h>

#include "ImageFile.h"



// Constructors, destructors, etc...
ImageFile::ImageFile(const char *filename, int read, int open) 
	: Image(),
	  _imageFileType(UNKNOWN_FILE),
	  __isOpen(0),
	  __drawWhileRead(1),
	  __filename(NULL),
	  __imageFile(NULL) {

	if (filename) {
		this->setFilename(filename);
	}

	if (open || read) {
		this->open();
	}
	if (read) {
		this->read();
	}
	if (open) {
		this->close();
	}
}


ImageFile::~ImageFile() {
	if (__isOpen) {
		this->close();
	}
	if (__filename) {
		delete [] __filename;
		__filename = NULL;
	}
}



void ImageFile::drawWhileRead(int state) {
	__drawWhileRead = state;
}


int ImageFile::setFilename(const char *filename) {

	if (__isOpen) {
		if (! this->close()) {
			return 0;
		}
	}

	if (__filename) {
		delete [] __filename;
		__filename = NULL;
	}

	int len = strlen(filename);
	__filename = new char[len+1];

	if (__filename) {
		strcpy(__filename, filename);
		__filename[len] = '\0';
		return 1;
	}
	else {
		return 0;
	}
}



const char* ImageFile::getFilename() {
	return __filename;
}




int ImageFile::open(const char* mode) {
	if (__isOpen) {
		this->close();
	}

	if (!__isOpen && __filename) {
		__imageFile = fopen(__filename, mode);
		if (__imageFile) {
			__isOpen = 1;
		}
		return __isOpen;
	}
	else {
		return 0;
	}
}



int ImageFile::open(const char *filename, const char* mode) {
	this->setFilename(filename);

	return this->open(mode);
}



int ImageFile::read()
{
	ImageFileType fileType;


	if (!__imageFile || !__isOpen) {
		if (!this->open("rb")) {
			return 0;
		}
	}


	if (_imageData) {
		delete [] _imageData;
		_imageData = NULL;
	}


	fileType = _findImageFileType();

	if (fileType == IMAGEFILE_ERROR) {
		return 0;
	}

	int retVal = 1;
	switch (fileType) {
		case RGB_FILE:
			_imageFileType = RGB_FILE;
			retVal &= _readRGBFile();
			break;

		case BMP_FILE:
			_imageFileType = BMP_FILE;
			retVal &= _readBMPFile();
			break;

		case GIF87A_FILE:
			_imageFileType = GIF87A_FILE;
			retVal &= _readGIF87aFile();
			break;

		default:
			retVal &= 0;
			break;
	}


	return retVal;
}




int ImageFile::read(const char *filename) {
	if (!this->open(filename, "rb")) {
		return 0;
	}
	
	return this->read();
}




int ImageFile::readDimensions() {

	ImageFileType fileType;


	if (!__imageFile || !__isOpen) {
		if (!this->open()) {
			return 0;
		}
	}

	fileType = _findImageFileType();

	if (fileType == IMAGEFILE_ERROR) {
		return 0;
	}

	switch (fileType) {
		case RGB_FILE:
			_imageFileType = RGB_FILE;
			return _readRGBHeader();

		case BMP_FILE:
			_imageFileType = BMP_FILE;
			return _readBMPHeader();

		case GIF87A_FILE:
			_imageFileType = GIF87A_FILE;
			return _readGIF87aHeader();

		default:
			return 0;
	}

}





int ImageFile::readDimensions(const char *filename) {
	if (!this->open(filename, "rb")) {
		return 0;
	}
	if (!readDimensions()) {
		this->close();
		return 0;
	}
	
	return this->close();
}





int ImageFile::write(ImageFileType type) {

	int retVal = 0;

	if (!__imageFile || !__isOpen) {
		if (!this->open("wb")) {
			return 0;
		}
	}


	switch(type) {
		case RGB_FILE:
			retVal = _writeRGBFile();
			break;

		case BMP_FILE:
			retVal = _writeBMPFile();
			break;

		case GIF87A_FILE:
			retVal = _writeGIF87aFile();
			break;

		default:
			break;
	}

	return retVal & this->close();
}





int ImageFile::write(const char *filename, ImageFileType type) {
	if (!this->open(filename, "wb")) {
		return 0;
	}

	return this->write(type);
}




int ImageFile::close() {
	int retVal = 0;

	if (__isOpen) {
		retVal = fclose(__imageFile);

		if (retVal == 0) {
			retVal = 1;
			__isOpen = 0;
			__imageFile = NULL;
		}
		else {
			retVal = 0;
		}
	}
	else {
		retVal = 1;
	}

	return retVal;
}






ImageFile::ImageFileType ImageFile::_findImageFileType() {
	if (!__imageFile || !__isOpen) {
		return IMAGEFILE_ERROR;
	}


	int status;
	char bytes[8];

	status = fseek(__imageFile, 0, SEEK_SET);

	if (status) {
		return IMAGEFILE_ERROR;
	}


	status = fread(bytes, 1, 8, __imageFile);

	if (status != 8) {
		return IMAGEFILE_ERROR;
	}



	//
	// see if this is a windows BMP file
	//
	if (strncmp(bytes, "BM", 2) == 0) {
		return BMP_FILE;
	}


	//
	// see if this is a GIF87a file....
	//
	if (strncmp(bytes, "GIF87a", 6) == 0) {
		return GIF87A_FILE;
	}


	//
	// see if we've got a rgb file
	//
	short magicNumber = ((unsigned short)((unsigned char)bytes[0])) << 8;
	magicNumber |= (unsigned short)((unsigned char)bytes[1]);

	if (magicNumber == 474) {
		return RGB_FILE;
	}


	return UNKNOWN_FILE;
}




