// This is ImageFileRGBWrite.cpp
//
// Copyright 1997, 1998     Mark Manca
//

#ifdef WIN32
#include <windows.h>
#endif

#include <GL/gl.h>
#include <string.h>
#include <stdio.h>

#include "ImageFile.h"
#include "rgbFileHeader.h"

#ifdef LITTLE_ENDIAN
#include "byteSwap.h"
#endif



int ImageFile::_writeRGBFile() {
	rgbFileHeader header;
	int status;


    if (!__imageFile || !__isOpen) {
        return 0;
    }

	//
	// get to the beginning of the file so we can write the
	// header out to the file
	//
	status = fseek(__imageFile, 0, SEEK_SET);
	if (status) {
		return 0;           // an error occured
	}


	//
	// get the header of the file all set up properly.
	//
	header.magicNumber = (short)474;
	header.storageFormat = (char)0;          // noRLE

	if (_imageDataType == GL_BYTE || _imageDataType == GL_UNSIGNED_BYTE) {
		header.bytesPerChannel = (char)1;    // used chars to represent data
	}
	else {
		header.bytesPerChannel = (char)2;    // used shorts 
	}

	header.numDimensions = 
		(unsigned short)(_numChannels == 1 ? 2 : 3);


	header.xSize = (unsigned short)_width;
	header.ySize = (unsigned short)_height;
	header.zSize = (unsigned short)_numChannels;

	if (_imageDataType == GL_BYTE) {
		header.minPixelValue = (long)-128;
		header.maxPixelValue = (long)127;
	}
	else if (_imageDataType == GL_UNSIGNED_BYTE) {
		header.minPixelValue = (long)0;
		header.maxPixelValue = (long)255;
	}
	else if (_imageDataType == GL_SHORT) {
		header.minPixelValue = (long)-32768;
		header.maxPixelValue = (long)32767;
	}
	else if (_imageDataType == GL_UNSIGNED_SHORT) {
		header.minPixelValue = (long)0;
		header.maxPixelValue = (long)65535;
	}

	memset(header.dummy, 0, 4);
	memset(header.imageName, 0, 80);
	header.colormapID = (long)0;
	memset(header.spaceFiller, 0, 404);


    #ifdef LITTLE_ENDIAN    // need to swap bytes
		header.magicNumber   = byteSwap(header.magicNumber);
		header.numDimensions = byteSwap(header.numDimensions);
		header.xSize         = byteSwap(header.xSize);
		header.ySize         = byteSwap(header.ySize);
		header.zSize         = byteSwap(header.zSize);
		header.minPixelValue = halfwordAndByteSwap(header.minPixelValue);
		header.maxPixelValue = halfwordAndByteSwap(header.maxPixelValue);
		header.colormapID    = halfwordAndByteSwap(header.colormapID);
    #endif

	// 
	// now that the header is set up, write it out to the file
	//
	status = fwrite((char*)&header, 1, sizeof(rgbFileHeader), __imageFile);
	if (status != sizeof(rgbFileHeader)) {
		return 0;
	}



	//
	// get some variables ready to go that we will use when we write
	// out the file
	//
	int rowDataSize;
	GLbyte *rowDataByte;
	GLubyte *rowDataUByte;
	GLshort *rowDataShort;
	GLushort *rowDataUShort;
	void *rowData;
	int dataMult;

	if (_imageDataType == GL_BYTE || _imageDataType == GL_UNSIGNED_BYTE) {
		dataMult = 1;       // is 1 because we are dealing w/chars
		rowDataSize = _width * dataMult;

		if (_imageDataType == GL_BYTE) {
			rowDataByte = new GLbyte[_width];
			rowData = rowDataByte;
		}
		else {       // _imageDataType == GL_UNSIGNED_BYTE
			rowDataUByte = new GLubyte[_width];
			rowData = rowDataUByte;
		}
	}
	else {      // we've got shorts...
		dataMult = 2;       // is 2 because we are dealing w/shorts
		rowDataSize = _width * dataMult;

		if (_imageDataType == GL_SHORT) {
			rowDataShort = new GLshort[_width];
			rowData = rowDataShort;
		}
		else {       // _imageDataType == GL_UNSIGNED_SHORT
			rowDataUShort = new GLushort[_width];
			rowData = rowDataUShort;
		}
    }



	//
	// now that the header and our extra data is ready to go, 
	// write out the file.
	//
	for (int i=0; i < _numChannels; i++) {
		for (int j=_height; j > 0; j--) {
			//
			// there is some code duplication here, but it is present
			// for efficiency's sake.
			//
			if (_imageDataType == GL_BYTE) {
				char* rowCharPtr = (char*)rowData;
				char* imageCharPtr = (char*)_imageData;

				imageCharPtr += ((_width*_numChannels + _extraRowWidth) * (_height - j)) + i;
				for (int k=_width; k > 0; k--) {
					*rowCharPtr++ = *imageCharPtr;
					imageCharPtr += _numChannels;
				}
			}
			else if (_imageDataType == GL_UNSIGNED_BYTE) {
				unsigned char* rowUCharPtr = (unsigned char*)rowData;
				unsigned char* imageUCharPtr = (unsigned char*)_imageData;

				imageUCharPtr += ((_width*_numChannels + _extraRowWidth) * (_height - j)) + i;
				for (int k=_width; k > 0; k--) {
					*rowUCharPtr++ = *imageUCharPtr;
					imageUCharPtr += _numChannels;
				}
			}
			else if (_imageDataType == GL_SHORT) {
				short* rowShortPtr = (short*)rowData;
				short* imageShortPtr = (short*)_imageData;

				imageShortPtr += ((_width*_numChannels + _extraRowWidth) * (_height - j)) + i;
				for (int k=_width; k > 0; k--) {
					*rowShortPtr = *imageShortPtr;

					#ifdef LITTLE_ENDIAN
					*rowShortPtr = byteSwap(*rowShortPtr);
					#endif

					rowShortPtr++;
					imageShortPtr += _numChannels;
				}
			}
			else if (_imageDataType == GL_UNSIGNED_SHORT) {
				unsigned short* rowUShortPtr = (unsigned short*)rowData;
				unsigned short* imageUShortPtr = (unsigned short*)_imageData;

				imageUShortPtr += ((_width*_numChannels + _extraRowWidth) * (_height - j)) + i;
				for (int k=_width; k > 0; k--) {
					*rowUShortPtr = *imageUShortPtr;

					#ifdef LITTLE_ENDIAN
					*rowUShortPtr = byteSwap(*rowUShortPtr);
					#endif

					rowUShortPtr++;
					imageUShortPtr += _numChannels;
				}
			}


			status = fwrite(rowData, 1, rowDataSize, __imageFile);
			if (status != rowDataSize) {
				return 0;
			}
		}
	}


	//
	// if we've made it this far, clean up and return 1
	//

	delete [] rowData;

	return 1;
}


