/******************************************************************
 * Intersect.c 
 *
 * The intersect subroutines 
 *
 *         Program #1 -- basic ray tracer
 ******************************************************************/

#include <stdio.h>
#include <math.h>
#include "SceneIO.h"
#include "Vector3f.h"
#include "Matrix4f.h"
#include "Intersect.h"

/* utility functions */
int max3 (float a, float b, float c);
float min (float a, float b);


/**************************************************************
 * Intersect()
 *
 * Finds the first intersect of a ray with a scene
 **************************************************************/
int Intersect (Ray *ray, SceneIO *scene, Isect *I, RTParameters* g_params)
{
	ObjIO *currObj;
	Isect currIntersect;
	float min_t;
	int num_hits;
	int min_num_hits;
	float dot;

	min_t = 100000;
	min_num_hits = 0;
	currObj = scene->objects;

	/* loop through all objects in the scene */
	while (currObj)
	{
		/* Sphere objects */	
		if (currObj->type == SPHERE_OBJ)
		{
			/* calc intersection point with sphere */	   
			if ( (num_hits = SphereIntersect (ray, currObj, &currIntersect, g_params) ) > 0)
			{
				/* keep it if this is the closest intersection so far */
				if ( (min_num_hits == 0) || (currIntersect.t < min_t) )
				{
					min_t = currIntersect.t;
					min_num_hits = num_hits;
					CopyIsect (&currIntersect, I);
				}
			}
		}

		/* Handle Polyset objects */
		if (currObj->type == POLYSET_OBJ)
		{
			/* Find closest intersection with polyset */
			if ( (num_hits = PolysetIntersect (ray, currObj, &currIntersect,g_params) ) > 0)
			{
				/* Keep this interesection if it is the closes so far */
				if ( (min_num_hits == 0) || (currIntersect.t < min_t) )
				{
					min_t = currIntersect.t;
					min_num_hits = num_hits;
					CopyIsect (&currIntersect, I);
				}
			}
		}

		currObj = currObj->next;
	}

	if ( min_num_hits == 0 ) 
		return 0;

	/* determine medium before and after intersection */
	dot = Vector3f_dot(ray->D, I->normal);
	if ( dot < 0 )
	{
		/* entering */
		I->prevIndex = 1.0;
		I->nextIndex = 1.5;
	}
	else
	{
		/* leaving */
		I->prevIndex = 1.5;
		I->nextIndex = 1.0;
	}

	/* return the number of intersections with the closest object */
	return min_num_hits;
}

/**************************************************************
 * SphereIntersect
 *
 * Calculates the nearest intersection of a ray with a sphere.
 * The total number of intersections is returned.
 **************************************************************/
int SphereIntersect (Ray *ray, ObjIO *object, Isect *intersect, RTParameters* g_params)
{
	Vector3f V;
	float b, disc;
	SphereIO *sphere;
	float t1, t2;
	Vector3f intersectPoint;

	sphere = (SphereIO *)object->data;

	Vector3f_sub (sphere->origin, ray->P, V);
	b = Vector3f_dot (V, ray->D);
	disc = b*b - Vector3f_dot (V, V) + sphere->radius*sphere->radius;
	if (disc < 0.)
	{
		return 0;
	}

	disc = sqrt(disc);
	t2 = b + disc;
	if (t2 <= g_params->rayeps)
	{
		return 0;
	}

	t1 = b - disc;

	if (t1 > g_params->rayeps)
	{
		/* entering */
		SetIsect(intersect, t1, object, 1, 1.0, 1.);

		RayPoint (ray, t1, intersectPoint);
		SphereNormal(object, intersectPoint, intersect->normal);

		return 2;
	} 
	else 
	{
		/* exitting */
		SetIsect(intersect, t2, object, 0, 1., 1.0);

		RayPoint (ray, t2, intersectPoint);
		SphereNormal(object, intersectPoint, intersect->normal);

		return 1;
	}
}

/**************************************************************
 * SphereNormal()
 *
 * Calculates the normal to a sphere at point P
 **************************************************************/
void SphereNormal (ObjIO *object, Vector3f P, Vector3f N)
{
	SphereIO *sphere;

	sphere = (SphereIO *)object->data;

	Vector3f_sub (P, sphere->origin, N);
	Vector3f_normalize(N);
}

/**************************************************************
 * PolysetIntersect()
 *
 * Calculates the first intersection of a ray with a polygon.
 **************************************************************/
int PolysetIntersect (Ray *ray, ObjIO *object, Isect *intersect, RTParameters* g_params)
{
	PolySetIO *polygon;
	int i;
	Vector3f normal;
	float min_t;
	float t;
	int num_hits;

	polygon = (PolySetIO *)object->data;
	min_t = 1e6;
	num_hits = 0;

	for (i=0; i < polygon->numPolys; i++)
	{
		if (TriangleIntersect(ray, &polygon->poly[i], &t, normal, g_params) )
		{
			/* hit! */
			if ( t < min_t )
			{
			min_t = t;
			Vector3f_copy(normal, intersect->normal);
			}
		num_hits++;
		}
	}  

	if (num_hits > 0)
	{
		SetIsect (intersect, min_t, object, 1, 1., 1.);
	}

	return num_hits;
}


/**************************************************************
 * TriangleIntersect()
 *
 * Computes the intersection of a ray with a triangle.
 * return 1 if hit, 0 if not.  Returns the t of the intersection
 * and the normal to the triangle
 **************************************************************/
int TriangleIntersect (Ray *ray, PolygonIO *triangle, float *t_int, Vector3f normal,
					   RTParameters* g_params)
{
	Vector3f edge1, edge2;
	float D;
	float t;
	Vector3f P;
	float u0, v0, u1, v1, u2, v2;
	float alpha, beta;
	int dominantAxis;
	int i0, i1, i2;
	float vd;

	Vector3f_sub (triangle->vert[1].pos, triangle->vert[0].pos, edge1);
	Vector3f_sub (triangle->vert[2].pos, triangle->vert[0].pos, edge2);

	Vector3f_cross (edge1, edge2, normal);
	Vector3f_normalize (normal);

	D = -( Vector3f_dot(normal, triangle->vert[1].pos) );

	vd = Vector3f_dot(normal, ray->D);
	if (vd == 0.) 
		return 0;	 

	t = -(Vector3f_dot(normal, ray->P) + D) / vd;

	if (t <= g_params->rayeps)
		return 0;

	RayPoint(ray, t, P);

	dominantAxis = max3(normal[0], normal[1], normal[2]);
	if (dominantAxis == 0)
	{
		i0=0;
		i1=1;
		i2=2;
	}
	else if (dominantAxis == 1)
	{
		i0=1;
		i1=0;
		i2=2;
	}
	else
	{
		i0=2;
		i1=0;
		i2=1;
	}

	u0 = P[i1] - triangle->vert[0].pos[i1];
	v0 = P[i2] - triangle->vert[0].pos[i2];

	u1 = triangle->vert[2].pos[i1] - triangle->vert[0].pos[i1];
	v1 = triangle->vert[2].pos[i2] - triangle->vert[0].pos[i2];

	u2 = triangle->vert[1].pos[i1] - triangle->vert[0].pos[i1];
	v2 = triangle->vert[1].pos[i2] - triangle->vert[0].pos[i2];

	alpha = -1;

	if (u1 == 0)
	{
		beta = u0/u2;
		if ( beta>=0 && beta <=1 )
			alpha = (v0 - beta*v2)/v1;
	}
	else
	{
		beta = (v0*u1 - u0*v1) / (v2*u1 - u2*v1);
		if ( beta>=0 && beta <=1 )
			alpha = (u0 - beta*u2)/u1;
	}

	*t_int = t;
	return ( (alpha >= 0) && (beta >= 0) && (alpha+beta <= 1) );
}


/**************************************************************
 * SetIsect()
 *
 * Fills the fields of an Isect structure.
 **************************************************************/
void SetIsect (Isect *I, float iT, ObjIO *iObject, int iEnter, float prevN, float nextN)
{
	I->t = iT;
	I->object = iObject;
	I->enter = iEnter;
	I->prevIndex = prevN;
	I->nextIndex = nextN;
}

/**************************************************************
 * CopyIsect()
 *
 * Copies an Isect structure.
 **************************************************************/
void CopyIsect (Isect *I1, Isect *I2)
{
	I2->t = I1->t;
	I2->object = I1->object;
	I2->enter = I1->enter;
	I2->prevIndex = I1->prevIndex;
	I2->nextIndex = I1->nextIndex;
	Vector3f_copy (I1->normal, I2->normal);
}

/**************************************************************
 * RayPoint()
 *
 * Finds a position on a ray at parameter t
 **************************************************************/
void RayPoint (Ray *ray, float t, Vector3f P)
{
	Vector3f temp;
  
	Vector3f_scale2 (t, ray->D, temp);
	Vector3f_add (ray->P, temp, P);
}

/**************************************************************
 * ClampColor()
 *
 * Clamps a color to (1,1,1)
 **************************************************************/
void ClampColor (Vector3f color)
{
	int i;

	for (i=0; i<3; i++)
	{
		if (color[i] > 1.0)
			color[i] = 1.0;
	}
}


/**************************************************************
 * max3()
 * 
 * Returns the index (0,1,2) of the maximum absolute value
 * of three numbers.
 **************************************************************/
int max3 (float a, float b, float c)
{
	a = abs(a);
	b = abs(b);
	c = abs(c);

	if ( a > b)
	{
		if (a > c) 
			return 0;
		else
			return 2;
	}
	else
	{
		if (b > c)
			return 1;
		else
			return 2;
	}
}

/**************************************************************
 * min()
 * 
 * Returns the minimum of 2 floats
 **************************************************************/
float min (float a, float b)
{
	if (a<b)
		return a;
	else
		return b;
}


