/********
*
*  SceneIO.h
*
*  Description:
*    Contains data structures for reading and writing scene descriptions.
*    The scene descriptions are originally from Inventor applications
*    (a toolkit provided by SGI), however rather than forcing you to
*    how to write an Inventor application, we convert them for you
*    to the much simpler data structures below.
*
*    Do not modify the data structures in this file.
*    You should define your own structures, and write a procedure that
*    that initializes them by walking through the structures defined here.
*
*  Authors: Brian L. Curless, Eric Veach
*  Date of Version: January 22, 1995
*
*********/

#ifndef _SCENE_IO_
#define _SCENE_IO_

#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif

#include <stdio.h>
#include "Vector3f.h"


    /* Definition of a perspective camera (projection/viewing) */

typedef struct _CameraIO
{
    Point3f  position;		/* Position of viewer				*/
    Vector3f viewDirection;	/* Viewing direction				*/
    float    focalDistance;	/* Distance from viewer to focal plane		*/
    Vector3f orthoUp;		/* Up direction, orthogonal to viewDirection	*/
    float    verticalFOV;	/* Total vertical field of view, in radians	*/
} CameraIO;


    /* We allow point, directional, and spot lights. */

enum LightType {POINT_LIGHT, DIRECTIONAL_LIGHT, SPOT_LIGHT};

typedef struct _LightIO
{
    _LightIO *next; /* Pointer to next light in linked list	*/
    enum LightType type;  /* Point, directional, or spot light		*/
    
    Point3f  position;	/* Position of a point light source		*/
    Vector3f direction;	/* Direction of a directional light source	*/
    Color3f  color;		/* Color and intensity				*/
    
    float	dropOffRate;	/* For spot lights, 0 <= x <= 1.  See man pages */
    float	cutOffAngle;	/* Angle at which spot light intensity is zero  */
} LightIO;



    /* Definition of material parameters assignable in Inventor		*/
    /*									*/
    /* Note that Inventor does not supply an index of refraction.	*/
    /* Of course, you can interpret each of these fields any way you	*/
    /* like.  For example, in a ray tracer the "emissive" and "ambient"	*/
    /* colors are pretty much the same thing, so you could interpret	*/
    /* one of them as something else.                                   */
    /*									*/
    /* If you need more parameters, you can hide them in the object 	*/
    /* name ("composer" and "i3dm" let you set this), and parse the	*/
    /* name as a string.						*/

typedef struct _MaterialIO
{
    Color3f diffColor;	/* Diffuse color				*/
    Color3f ambColor;	/* Ambient color				*/
    Color3f specColor;	/* Specular color				*/
    Color3f emissColor;	/* Emissive color				*/
    float	shininess;	/* Shininess - must be scaled before use as a   */
						/*   Phong cosine exponent: 0.0 - 1.0	        */
    float	ktran;		/* Transparency: 0.0 - 1.0			*/
} MaterialIO;




    /* Definition of a generic object.		*/
    /* We support spheres and sets of polygons. */

enum ObjType {SPHERE_OBJ, POLYSET_OBJ};

typedef struct _ObjIO
{
    _ObjIO *next;		/* Pointer to next object node	    */
    char *name;                 /* Name of the object               */

    /* Most objects are made out of one material.  However, there */
    /* may be several materials for one object if per-vertex materials */
    /* are used.  Each vertex gives an index into the material list. */
    /* "material" is a pointer to an array of MaterialIO structures */
    /* (although often this array will have length 1). */

    long numMaterials;
    MaterialIO *material;
    
    enum ObjType type;		/* See enum ObjType		    */
    void *data;			/* data for this type (see below)   */
} ObjIO;



/* Definition of the root of the scene description */

typedef struct _SceneIO
{
    CameraIO *camera;  /* Perspective camera		      */
    LightIO *lights;   /* Head of the linked list of lights    */
    ObjIO *objects;    /* Head of the linked list of objects   */
} SceneIO;



    /* Definition of a spherical object.			    */

    /* In fact, due to the ability to scale the axes		    */
    /* independently, spheres can be changed to ellipsoids.	    */
    /* Accordingly, the axis orientations are defined here	    */
    /* as well as their lengths.  The "radius" is defined here      */
    /* for convenience as half the average axis length, which is    */
    /* exact for spheres.					    */

typedef struct _SphereIO
{
    Point3f origin;		/* Origin of the sphere/ellipsoid		*/
    float	radius;	    /* Half the sum of the axis lengths		*/
    
    Vector3f xaxis;	    /* Direction of the x axis			*/
    float	 xlength;   /* Length of the x axis			*/
    Vector3f yaxis;	    /* Direction of the y axis			*/
    float	 ylength;   /* Length of the y axis			*/
    Vector3f zaxis;	    /* Direction of the z axis			*/
    float	 zlength;   /* Length of the z axis			*/
} SphereIO;



/* Definition of a vertex */

typedef struct _VertexIO
{
    Point3f  pos;			/* Vertex position */
    Vector3f norm;			/* Normal to vertex (if PER_VERTEX_NORMAL) */
    long materialIndex;		/* Index into materials list (see ObjIO), */
                            /* if material is PER_VERTEX_MATERIAL */
    float s, t;				/* Texture coordinates (if hasTextureCoords) */
} VertexIO;


    /* Definition of a polygon */

typedef struct _PolygonIO {
    long numVertices;		/* Number of vertices		*/
    VertexIO *vert;			/* Vertex array				*/
} PolygonIO;



    /* Definition of a polygonal set.									*/
    /* We support 3 types:												*/
    /*																	*/
    /*   Triangle mesh --												*/
    /*     Consists only of triangles.  There is not necessarily any	*/
    /*	   structure to the triangles; although the triangles will		*/
    /*     usually be from a single modelling primitive.				*/
    /*																	*/
    /*   Face set --													*/
    /*	   A set of polygons with any number of sides.  They will		*/
    /*	   always have per-face normals and per-object materials.		*/
    /*																	*/
    /*   Quad mesh --													*/
    /*	   All "polygons" have four sides, but are NOT necessarily		*/
    /*     planar, i.e. the 4 vertices of each quad may not lie in		*/
    /*     the same plane.												*/

enum PolySetType {POLYSET_TRI_MESH, POLYSET_FACE_SET, POLYSET_QUAD_MESH};


    /* Objects (polygonal sets in particular) may have normals */
    /*   at every vertex for smooth shading of meshed objects  */
    /*   or at every face for sharp-edged shading as in boxes  */

enum NormType {PER_VERTEX_NORMAL, PER_FACE_NORMAL};


    /* We support two kinds of material bindings:					*/
    /*	PER_OBJECT    one material for the whole object				*/
    /*  PER_VERTEX    a material is defined at each vertex, and you	*/
    /*			must interpolate across each face.					*/
    /*																*/
    /* PER_VERTEX obviously only applies to polygonal objects.		*/

enum MaterialBinding {PER_OBJECT_MATERIAL, PER_VERTEX_MATERIAL};


typedef struct _PolySetIO
{
    enum PolySetType		type;
    enum NormType		normType;
    enum MaterialBinding	materialBinding;
    int /* boolean */		hasTextureCoords;
    
    /* Sometimes it may be useful to know when the polygonal set is
     * organized into a rectangular pattern.  If the primitives are
     * are grouped into rows of equal length, "rowSize" gives the
     * number of primitives in each row, otherwise "rowSize" is 0.
     * The number of rows is (numPolys / rowSize), an integer.
     */
    long rowSize;     	    /* Number of primitives per row		*/
    long numPolys;			/* Number of polygons				*/
    PolygonIO *poly; /* Polygonal array					*/
} PolySetIO;




/* SceneIO *read_scene(filename)
 *    - loads a scene description previously saved by the "composer" tool.
 *
 * SceneIO *request_composer_scene()
 *    - sends a message to "composer" asking for the current scene.  If
 *      "composer" is running, it exports the current scene and supplies
 *      the name of the intermediate file.
 *
 * void delete_scene(scene)
 *    - call this when you are finished with a scene returned by
 *      read_scene() or request_composer_scene().
 */

SceneIO *read_scene(const char *);
SceneIO *request_composer_scene(void);
void delete_scene(SceneIO *);


/* The following routines are used to construct new scenes.  They are
 * used by "composer" when translating Inventor files.  You should not
 * need to access them.
 */

SceneIO *new_scene(void);
CameraIO *new_camera(void);
LightIO *new_light(void);
ObjIO *new_object(void);
MaterialIO *new_material(long n);

LightIO *append_light(LightIO **);
ObjIO *append_object(ObjIO **);

/* write_scene_ascii() writes a scene in an ASCII format.  This format is
 * useful for debugging, or transferring scene files between different
 * machine architectures.  By writing to the file "/dev/tty" you can
 * write to the terminal window in which your application was invoked.
 *
 * write_scene_binary() writes a scene in binary format; the Export function
 * in composer uses this format to write the scene.  This format can only
 * be read back on a machine that uses the same format for integers
 * and floating-point numbers.
 */
void write_scene_ascii(SceneIO *, const char *);
void write_scene_binary(SceneIO *, const char *);


/* For interaction with the "composer" tool */

#define COMPOSER_EXPORT_REQUESTED	0
#define COMPOSER_EXPORT_SENDING		1
#define COMPOSER_EXPORT_DONE		2

#define COMPOSER_DEFAULT_EXPORT_NAME	"composer.out"


#endif


